/*
 * $Rcsid$
 * 
 * promisc.c -- Solaris kernel based promiscuous interface detection.
 *              A white hat early warning system. Apply liberally with
 *              msyslogd and anonymous routing. Lather. Rinse. Repeat.
 *
 * User Datagram Protocol <udp@low-level.net>
 * 26-Nov-2000
 *
 * 95% functional. Seems stable. Only tested on sun4m/Solaris 2.6.
 * This code won't detect a kernel based sniffer as should be self evident,
 * so I guess that's a subject for future code.
 * The code appears to be 64-bit clean, however, there are thunking issues
 * on Solaris 8/64.
 *
 * ISSUE: Can't detect interface going *out* of promiscuous mode, unless
 * we trap and follow closes of STREAMS which reference the same hardware.
 * This means that you can use this as your syslog() early warning system
 * for the kiddies, but you can't use it to get precise metrics about who
 * sniffed when and for how long. Most DLPI clients simply close the stream
 * rather than issue a DL_PROMISCOFF_REQ.
 * 
 * TODO: Lookup interface names automagically for output.
 *  - Gone some way towards this by looking up the driver name from major.
 *    Minor always seems to be 2 for sun4m onboard le - wtf? should map directly
 *    to an instance number.
 * TODO: Walk ill_g_head and frob the ill_t flags with IFF_PROMISC.
 * TODO: Maintain a kstat for each interface with timestamps (when
 *       DL_PROMISCON happened, and for how long).
 * TODO: Make the module resistant to tampering.
 *
 */
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/cmn_err.h>
#include <sys/modctl.h>
#include <sys/open.h>
#include <sys/kstat.h>
#include <sys/systm.h>
#include <sys/param.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/buf.h>
#include <sys/modctl.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/cred.h>
#include <sys/dditypes.h>
#include <sys/stream.h>
#include <sys/poll.h>
#include <sys/autoconf.h>
#include <sys/byteorder.h>
#include <sys/socket.h>
#include <sys/dlpi.h>
#include <sys/stropts.h>
#include <sys/sockio.h>
#include <sys/syscall.h>

#include <sys/ddi.h>
#include <sys/sunddi.h>

#include <net/if.h>
#include <inet/common.h>
#include <inet/led.h>
#include <inet/ip.h>

/*
 * this could be modlsys, but that's not quite as much fun.
 * modlsys is the officially supported way of doing this - from
 * within mod_install(), the kernel will call mod_installsys()
 * on our behalf, which does actually use rw_enter(), but I'm not
 * sure which kernel execution context is used to do this.
 *
 */
static	struct modlmisc modlmisc = {
	&mod_miscops,
	"ews 0.1"
};
static	struct modlinkage modlinkage = {
	MODREV_1,
	{&modlmisc,NULL,NULL,NULL}
};

/* we require kstat, for the maintenance of kstats. */
/*char _depends_on[] = "drv/kstat";*/

static void print_function(char *string){
	cmn_err(CE_WARN, "promisc: In routine %s", string);
}

/*--------------------------------------------------------------------*/

/* both rval and c calling conventions appear to be in use by putmsg(). */

typedef struct putmsg_args {
	int fd;
	struct strbuf *cp;
	struct strbuf *dp;
	int flags;
} putmsg_args_t;

static int promisc_syscall_putmsg(int fd, struct strbuf *cp, struct strbuf *dp,
	int flags);
static int promisc_syscall_rval_putmsg(struct putmsg_args *argp, rval_t *rvalp);
typedef int (*putmsg_rval_fn_t)(struct putmsg_args *argp, rval_t *rvalp);
typedef int (*putmsg_fn_t)(int fd, struct strbuf *cp, struct strbuf *dp,
        int flags); 

/*--------------------------------------------------------------------*/

/* cache old syscall details */
static struct sysent old_sysent_putmsg;

/*--------------------------------------------------------------------*/

int _init(void)
{
       	int r = -1;

	print_function("_init");
	if ((r = mod_install(&modlinkage)) != 0) {
		print_function("init: modinstall != 0");
		cmn_err(CE_WARN, "promisc: _init FAILED");
		return r;
	}

	cmn_err(CE_CONT, "promisc: saving and redirecting putmsg() syscall\n");
	bcopy(&sysent[SYS_putmsg], &old_sysent_putmsg, sizeof(struct sysent));

	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_call was: %p\n", sysent[SYS_putmsg].sy_call);
	sysent[SYS_putmsg].sy_call = (void *) promisc_syscall_rval_putmsg;
	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_call is now: %p\n", sysent[SYS_putmsg].sy_call);
	/* only replace the C one */

	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_callc was: %p\n", sysent[SYS_putmsg].sy_callc);
	sysent[SYS_putmsg].sy_callc = (void *) promisc_syscall_putmsg;
	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_callc is now: %p\n", sysent[SYS_putmsg].sy_callc);

	cmn_err(CE_CONT, "promisc: _init returns SUCCESS.\n");
	return r;
}

int _info(struct modinfo *pinfo)
{
	/* this is crashing out for some reason */
	cmn_err(CE_CONT, "promisc: _info called.");
	return (mod_info(&modlinkage, pinfo));
}

int _fini(void)
{
	int r;

	cmn_err(CE_CONT, "promisc: _fini called\n");
	cmn_err(CE_CONT, "promisc: restoring old syscall\n");
	sysent[SYS_putmsg].sy_call = old_sysent_putmsg.sy_call;
	sysent[SYS_putmsg].sy_callc = old_sysent_putmsg.sy_callc;

	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_callc is now: %p\n", sysent[SYS_putmsg].sy_callc);
	cmn_err(CE_CONT, "promisc: sysent[SYS_putmsg].sy_call is now: %p\n", sysent[SYS_putmsg].sy_call);

	r = mod_remove(&modlinkage);
	return r;
}

/*--------------------------------------------------------------------*/

static int promisc_syscall_putmsg(
	int fd,
	struct strbuf *cp,
	struct strbuf *dp,
        int flags)
{
	int r;

	/* doesn't appear to get called much - merely seems to be an
	 * interface to autopush. may need to override this properly on
 	 * systems which have a firewall (other than ipf) installed.
	 */
	r = ((putmsg_fn_t)old_sysent_putmsg.sy_callc)(fd, cp, dp, flags);

	return r;
}

static int promisc_syscall_rval_putmsg(struct putmsg_args *argp, rval_t *rvalp)
{
	struct strbuf *kcp = NULL;
	dl_promiscon_req_t *krt = NULL;
	int r = EINVAL;

	/* copy the strbuf into kernel space for brief inspection */
	kcp = kmem_alloc(sizeof(struct strbuf), KM_SLEEP);	
	copyin((void *)argp->cp, (void *)kcp, sizeof(struct strbuf));
	krt = kmem_alloc(sizeof(dl_promiscon_req_t), KM_SLEEP);
	copyin((void *)kcp->buf, (void *)krt, sizeof(dl_promiscon_req_t));

	/* dump promisc status to kernel dmesg buffer */
	if (((krt->dl_primitive == DL_PROMISCON_REQ)
	    || (krt->dl_primitive == DL_PROMISCOFF_REQ))
	    && (krt->dl_level == DL_PROMISC_PHYS))
	{
		file_t 	*fp;
		vnode_t *vp;
		struct cred	*cr;
		dev_t	rd;
		struct stdata *st;
		char *s;

		/* get file info */	
		if ((fp = GETF(argp->fd)) == NULL) {
			r = set_errno(EBADF);
			goto do_free;
		}
		vp = fp->f_vnode;
		cr = fp->f_cred; 
		rd = vp->v_rdev;
		if (krt->dl_primitive == DL_PROMISCON_REQ) {
			cmn_err(CE_WARN, "%s: promiscuous mode enabled (uid=%d gid=%d)",
			((s = ddi_major_to_name(getmajor(rd))) ? s : "unknown"),
			cr->cr_ruid, cr->cr_rgid );
		} else {
			cmn_err(CE_WARN, "promiscuous mode disabled");
		}
		RELEASEF(argp->fd);
	}

	/* call previous putmsg() routine */
	r = ((putmsg_rval_fn_t)old_sysent_putmsg.sy_call)(argp, rvalp);

do_free:
	kmem_free(krt, sizeof(dl_promiscon_req_t));
	kmem_free(kcp, sizeof(struct strbuf));
	return r;
}

/*--------------------------------------------------------------------*/
